/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "client.h"

/* globals */

#define POOL_SIZE 4

u32 __shutdown_flag;
u64 __speed;
v3_CONFIG __config;
PACKET_ACTIVE __work_pool[POOL_SIZE];

void catch_signal( int arg )
{
  __shutdown_flag = 1;
  v3SignalRegister( V3_PROCESS_SIGINT, catch_signal );
  v3SignalRegister( V3_PROCESS_SIGTERM, catch_signal );
}

u32 GetWork( PACKET_ACTIVE * work, u32 count )
{
  u32 i, length, received;
  v3_NET net;
  PACKET_REQUEST request;
  PACKET_ASSIGNMENT assign;

  received = 0;

  /* open connection to server */
  v3MemSet( &net, v3u64u32( sizeof( v3_NET ) ), 0 );
  if ( Connect( &net ) != V3_PASS )
  {
    return( received );
  }

  /* save network data */
  request.type = PACKET_TYPE_REQUEST;
  v3u32Save( (u8 *) &request.type, &request.type );
  v3u64Save( (u8 *) &request.speed, &__speed );

  /* get up to count assignments */
  for ( i = 0 ; i < count ; i++ )
  {
    if ( work[i].state != STATE_INVALID )
    {
      /* dont overwrite good data, skip to next one */
      continue;
    }

    /* send request */
    if ( v3NetSend( &net, &request, (u32) sizeof( PACKET_REQUEST ) )
      != V3_PASS )
    {
      v3PrintA( (ascii *) "Network send failure\n" );
      v3NetClose( &net );
      return( received );
    }

    /* read assignment */
    length = (u32) sizeof( PACKET_ASSIGNMENT );
    /* make sure to wait long enough, about 1sec/KB + 5sec */
    if ( v3NetRecv( &assign, length, &net, length + 5000 ) != length )
    {
      v3PrintA( (ascii *) "Network Recv Timeout\n" );
      v3NetClose( &net );
      return( received );
    }

    /* load network data */
    v3u32Load( &assign.type, (u8 *) &assign.type );
    v3u32Load( &assign.start, (u8 *) &assign.start );
    v3u32Load( &assign.end, (u8 *) &assign.end );

    if ( assign.type != PACKET_TYPE_ASSIGNMENT )
    {
      v3NetClose( &net );
      v3PrintA( (ascii *) "Bad block type\n" );
      return( received );
    }

    if ( assign.end == 0 )
    {
      v3NetClose( &net );
      v3PrintA( (ascii *) "Server out of work\n" );
      return( received );
    }

    /* copy into a PACKET_ACTIVE - the Runable data type */
    work[i].type = PACKET_TYPE_ACTIVE;
    work[i].start = assign.start;
    work[i].current = assign.start;
    work[i].end = assign.end;
    work[i].total = v3u64u32( 0 );
    work[i].state = STATE_ALIVE;

    v3PrintA( (ascii *) "Receieved assignment block %u/%u\n",
      received + 1, count );
    v3PrintA( (ascii *) "current %08X / end %08X\n",
      work[i].current, work[i].end );

    received++;
  }

  v3NetClose( &net );
  return( received );
}

u32 PutWork( PACKET_ACTIVE * work, u32 count )
{
  u32 i, length, sent;
  v3_NET net;
  PACKET_RESULTS results;
  PACKET_ACCEPT work_accept;

  sent = 0;

  /* open connection to server */
  v3MemSet( &net, v3u64u32( sizeof( v3_NET ) ), 0 );
  if ( Connect( &net ) != V3_PASS )
  {
    return( sent );
  }

  /* get up to count assignments */
  for ( i = 0 ; i < count ; i++ )
  {
    if ( work[i].state != STATE_DONE )
    {
      /* it's not done, done send this one back, move on to next one */
      continue;
    }

    /* save network data */
    results.type = PACKET_TYPE_RESULTS;
    v3u32Save( (u8 *) &results.type, &results.type );
    v3u32Save( (u8 *) &results.start, &work[i].start );
    v3u32Save( (u8 *) &results.end, &work[i].end );
    v3u64Save( (u8 *) &results.total, &work[i].total );
    results.cpu = CPU_TYPE;
    v3u32Save( (u8 *) &results.cpu, &results.cpu );
    results.os = OS_TYPE;
    v3u32Save( (u8 *) &results.os, &results.os );
    v3StrCopyA( results.email, v3ConfigGet( &__config, (ascii *) "settings",
      (ascii *) "email" ), v3u64u32( sizeof( results.email ) ) );

    /* send request */
    if ( v3NetSend( &net, &results, (u32) sizeof( PACKET_RESULTS ) )
      != V3_PASS )
    {
      v3PrintA( (ascii *) "Network send failure\n" );
      v3NetClose( &net );
      return( sent );
    }

    /* read assignment */
    length = (u32) sizeof( PACKET_ACCEPT );
    /* make sure to wait long enough, about 1sec/KB + 5sec */
    if ( v3NetRecv( &work_accept, length, &net, length + 5000 ) != length )
    {
      v3PrintA( (ascii *) "Network Recv Timeout\n" );
      v3NetClose( &net );
      return( sent );
    }

    /* load network data */
    v3u32Load( &work_accept.type, (u8 *) &work_accept.type );
    v3u32Load( &work_accept.result, (u8 *) &work_accept.result );

    if ( work_accept.type != PACKET_TYPE_ACCEPT )
    {
      v3NetClose( &net );
      v3PrintA( (ascii *) "Bad block type\n" );
      return( sent );
    }

    /* if work_accept.result == V3_PASS */
    v3PrintA( (ascii *) "Sent result block %u/%u\n", sent + 1, count );
    v3PrintA( (ascii *) "start %08X / end %08X\n",
      work[i].start, work[i].end );

    work[i].state = STATE_INVALID;

    sent++;
  }

  v3NetClose( &net );
  return( sent );
}

s32 Connect( v3_NET * net )
{
  u32 ip, fire_ip, port, fire_port;
  s32 result;

  if ( v3NetDNS( &ip, 1, (ascii *) v3ConfigGet( &__config,
    (ascii *) "settings", (ascii *) "server" ) ) != 1 )
  {
    v3PrintA( (ascii *) "Hostname lookup error.\n" );
    return( V3_FAIL );
  }

  if ( v3u32A( &port, NULL, v3ConfigGet( &__config, (ascii *) "settings",
    (ascii *) "port" ), 10 ) != V3_PASS )
  {
    v3PrintA( (ascii *) "Port error.\n" );
    return( V3_FAIL );
  }

  /* check is firewall is active */
  if ( v3StrCmpA( (ascii *) "yes", v3ConfigGet( &__config,
    (ascii *) "firewall", (ascii *) "active" ), v3u64u32( 3 ) ) == 0 )
  {
    if ( v3NetDNS( &fire_ip, 1, (ascii *) v3ConfigGet( &__config,
      (ascii *) "firewall", (ascii *) "host" ) ) != 1 )
    {
      v3PrintA( (ascii *) "Firewall hostname lookup error.\n" );
      return( V3_FAIL );
    }

    if ( v3u32A( &fire_port, NULL, v3ConfigGet( &__config,
      (ascii *) "firewall", (ascii *) "port" ), 10 ) != V3_PASS )
    {
      v3PrintA( (ascii *) "Firewall port error.\n" );
      return( V3_FAIL );
    }

   v3PrintA( (ascii *) "Connect: %u.%u.%u.%u/%u - fire %u.%u.%u.%u/%u\n",
    ( ip >> 24 ) & 0xFF, ( ip >> 16 ) & 0xFF,
    ( ip >> 8 ) & 0xFF, ip & 0xFF, port,
    ( fire_ip >> 24 ) & 0xFF, ( fire_ip >> 16 ) & 0xFF,
    ( fire_ip >> 8 ) & 0xFF, fire_ip & 0xFF, fire_port );

    result = v3NetOpen( net, ip, (u16) port, 0, 0, V3_NET_MODE_TCP,
      fire_ip, (u16) fire_port, (ascii *) v3ConfigGet( &__config,
      (ascii *) "firewall", (ascii *) "userpass" ) );
  }
  else
  {
   v3PrintA( (ascii *) "Connect: %u.%u.%u.%u port %u\n",
    ( ip >> 24 ) & 0xFF, ( ip >> 16 ) & 0xFF,
    ( ip >> 8 ) & 0xFF, ip & 0xFF, port );

    result = v3NetOpen( net, ip, (u16) port, 0, 0, V3_NET_MODE_TCP,
      0, 0, NULL );
  }

  if ( result != V3_PASS )
  {
    switch ( result )
    {
      case V3_NET_ERROR_HOST:
      case V3_NET_ERROR_PORT:
        v3PrintA( (ascii *) "Network setting error.\n" );
        break;
      case V3_NET_ERROR_FIREHOST:
      case V3_NET_ERROR_FIREPORT:
      case V3_NET_ERROR_FIREPASS:
        v3PrintA( (ascii *) "Firewall setting error.\n" );
        break;
      case V3_NET_ERROR_CLOSED:
      case V3_NET_ERROR_NO_NET:
        v3PrintA( (ascii *) "System unable to open network connection.\n" );
        break;
      default:
        v3PrintA( (ascii *) "Error opening network connection.\n" );
        break;
    }
    return( V3_FAIL );
  }
  return( V3_PASS );
}

u32 Load( void )
{
  /* Load in all PACKET_ACTIVE's from CLIENT_DATAFILE */
  /* Load all data into correct format using v3{type}Load() */

  return( 0 );
}

s32 Save( void )
{
  /* Save all PACKET_ACTIVE's to CLIENT_DATAFILE */
  /* Save all data into correct format using v3{type}Save() */

  return( V3_PASS );
}

s32 Run( PACKET_ACTIVE * work, u64 iterations )
{
  u64 loops;

  if ( work->type != PACKET_TYPE_ACTIVE )
  {
    return( STATE_INVALID );
  }

  if ( work->state != STATE_ALIVE )
  {
    return( STATE_DONE );
  }

  _V3_SET64( loops, 00000000, 00000000 );

  /* while work to do, not interupted, and not over our iteration limit */
  while ( ( work->current <= work->end ) && ( __shutdown_flag == 0 ) &&
    ( v3u64Lt( loops, iterations ) ) )
  {
    work->total = v3u64Add( work->total, v3u64u32( work->current ) );
    work->current++;

    v3u64Inc( &loops );
  }

  /* if out of work to do */
  if ( work->current > work->end )
  {
    work->state = STATE_DONE;
  }

  return( STATE_ALIVE );
}

u64 Speed( void )
{
  PACKET_ACTIVE test;
  v3_time start, end, diff, span;
  u128 bigcount, udiff;
  u64 count, limit;

  v3PrintA( (ascii *) "Benchmarking " );

  /* default speed */
  count = v3u64u32( 0x10000 );

  /*
    benckmark over a 1 second span, this will limit our resolution.
    test will take max of 2 x span.
    If the work being done takes a large time relative to a second, then
    you will need to benchmark over a longer period to get the needed
    resolution.
  */
  _V3_SET128( span, 00000000, 00000001, 00000000, 00000000 );

  _V3_SET64( limit, FFFFFFFF, FFFFFFFF );
  do
  {
    v3PrintA( (ascii *) "." );
    count = v3u64Lsh( count, 1 );
    test.type = PACKET_TYPE_ACTIVE;
    test.current = 0;
    test.end = v3u32u64( count );
    test.total = v3u64u32( 0 );
    test.state = STATE_ALIVE;
/* !!! need to deal with error cases of clock or Run() */
    v3SystemClock( &start );
    Run( &test, limit );
    v3SystemClock( &end );
    diff = v3s128Sub( end, start );
  } while ( v3s128Lt( diff, span ) && ( __shutdown_flag == 0 ) );

  /* make an unsigned udiff from diff */
  udiff = v3u128s128( diff );

  /*
    calculate how many per unit time, this determines how much work
    we request, change the 1 below to a 86400 or 1 day of work at a time.
    It's safe to assume that if we can do 2^64 units in a day, then we sure
    do not need client-server to do it :)
  */
  bigcount.hi = v3u64Mul( count, v3u64u32( 1 /* seconds */ ) );
  bigcount.lo = v3u64u32( 0 );
  bigcount = v3u128Div( bigcount, udiff );
  count = bigcount.lo;

  /* make sure it's more then 0 */
  if ( v3u64Eq( count, v3u64u32( 0 ) ) )
  {
    v3u64Inc( &count );
  }

  v3PrintA( (ascii *) "\nSpeed = %v\n", count );

  return( count );
}

s32 Test( void )
{
  PACKET_ACTIVE test;
  u64 correct, limit;
  s32 err;

  /* Run with some test data, be sure it's more then one Run() worth */
  test.type = PACKET_TYPE_ACTIVE;
  test.current = 0;
  test.end = 0xFFFF;
  test.total = v3u64u32( 0 );
  test.state = STATE_ALIVE;
  _V3_SET64( limit, 00000000, 00001000 );
  do
  {
    err = Run( &test, limit );
  } while ( ( err == STATE_ALIVE ) && ( __shutdown_flag == 0 ) );

  _V3_SET64( correct, 00000000, 7FFF8000 );
  if ( !v3u64Eq( test.total, correct ) )
  {
    return( -1 );
  }

  /* need more tests */

  return( V3_PASS );
}

s32 Config( void )
{
  ascii buffer[64];
  u64 length;

  length = v3u64u32( 64 );

  v3PrintA( (ascii *) "Email/ID [%.63s]: ", v3ConfigGet( &__config,
    (ascii *) "settings", (ascii *) "email" ) );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
  {
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "email", buffer );
    v3ConfigSave( &__config, CLIENT_CONFIG );
  }

  v3PrintA( (ascii *) "Server [%.63s]: ", v3ConfigGet( &__config,
    (ascii *) "settings", (ascii *) "server" ) );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
  {
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "server",
      buffer );
    v3ConfigSave( &__config, CLIENT_CONFIG );
  }

  v3PrintA( (ascii *) "Port [%.5s]: ", v3ConfigGet( &__config,
    (ascii *) "settings", (ascii *) "port" ) );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
  {
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "port", buffer );
    v3ConfigSave( &__config, CLIENT_CONFIG );
  }

  v3PrintA( (ascii *) "Logging level [%.5s]: ", v3ConfigGet( &__config,
    (ascii *) "settings", (ascii *) "loglevel" ) );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
  {
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "loglevel",
      buffer );
    v3ConfigSave( &__config, CLIENT_CONFIG );
  }

  v3PrintA( (ascii *) "Firewall yes/no [%.63s]: ", v3ConfigGet( &__config,
    (ascii *) "firewall", (ascii *) "active" ) );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
  {
    /* clean input */
    if ( ( buffer[0] == 'y' ) || ( buffer[0] == 'Y' ) )
    {
      v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "active",
        (ascii *) "yes" );
    }
    else
    {
      v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "active",
        (ascii *) "no" );
    }
    v3ConfigSave( &__config, CLIENT_CONFIG );
  }

  if ( v3StrCmpA( v3ConfigGet( &__config, (ascii *) "firewall",
    (ascii *) "active" ), (ascii *) "yes", v3u64u32( 3 ) ) == 0 )
  {
    v3PrintA( (ascii *) "Firewall [%.63s]: ", v3ConfigGet( &__config,
      (ascii *) "firewall", (ascii *) "host" ) );
    v3InputA( buffer, length, V3_IO_ECHO );
    if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
    {
      v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "host",
        buffer );
      v3ConfigSave( &__config, CLIENT_CONFIG );
    }

    v3PrintA( (ascii *) "Firewall Port [%.63s]: ", v3ConfigGet( &__config,
      (ascii *) "firewall", (ascii *) "port" ) );
    v3InputA( buffer, length, V3_IO_ECHO );
    if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
    {
      v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "port",
        buffer );
      v3ConfigSave( &__config, CLIENT_CONFIG );
    }

    v3PrintA( (ascii *) "Firewall user:pass [%.63s]: ", v3ConfigGet(
      &__config, (ascii *) "firewall", (ascii *) "userpass" ) );
    v3InputA( buffer, length, V3_IO_ECHO );
    if ( v3u64Gt( v3StrLengthA( buffer ), v3u64u32( 0 ) ) )
    {
      v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "userpass",
        buffer );
      v3ConfigSave( &__config, CLIENT_CONFIG );
    }
  }

  return( V3_PASS );
}

int main( int argc, char * argv[] )
{
  u32 i;
  u32 count;
  s32 err, err2;
  u64 limit;

  /* run system tests */
  v3PrintA( (ascii *) "Running system tests... " );
  if ( v3Test( &err, &err2, 0 ) != V3_PASS )
  {
    v3PrintA( (ascii *) "\nPlease report error: O=%u, C=%u, M=%i, E=%i\n",
      OS_TYPE, CPU_TYPE, err, err2 );
    v3ProcessEnd( -1 );
  }
  v3PrintA( (ascii *) "all passed.\n" );

  /* initialize the globals */
  v3MemSet( &__config, v3u64u32( sizeof( v3_CONFIG ) ), 0 );
  v3MemSet( &__work_pool,
    v3u64u32( sizeof( PACKET_ACTIVE ) * POOL_SIZE ), 0 );

  /* set client to lowest priority */
  v3ProcessPriority( 1 );

  /* register signals, after this we have to check the flag for shutdown */
  __shutdown_flag = 0;
  if ( ( v3SignalRegister( V3_PROCESS_SIGINT, catch_signal ) != V3_PASS )
    || ( v3SignalRegister( V3_PROCESS_SIGTERM, catch_signal ) != V3_PASS ) )
  {
    v3PrintA( (ascii *) "Signal registration failed\n" );
    v3ProcessEnd( -1 );
  }

  v3PrintA( (ascii *) "Running self-tests... " );
  if ( ( err = Test() ) != V3_PASS )
  {
    v3PrintA( (ascii *) "Please report error: O=%u, C=%u, P=%i\n",
      OS_TYPE, CPU_TYPE, err );
    v3ProcessEnd( -1 );
  }
  v3PrintA( (ascii *) "all passed.\n", err );

  /* load and check __config */
  v3ConfigLoad( &__config, CLIENT_CONFIG );
  if ( v3ConfigGet( &__config, (ascii *) "settings", (ascii *) "email" )
    == NULL )
  {
    /* set defaults */
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "email",
      (ascii *) "cs-sdk" );
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "server",
      NETPACKET_SERVER );
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "port",
      NETPACKET_PORT );
    v3ConfigSet( &__config, (ascii *) "settings", (ascii *) "loglevel",
      (ascii *) "0" );
    v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "active",
      (ascii *) "no" );
    v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "host",
      (ascii *) "hostname" );
    v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "port",
      (ascii *) "1080" );
    v3ConfigSet( &__config, (ascii *) "firewall", (ascii *) "userpass",
      (ascii *) "user:password" );
    v3ConfigSave( &__config, CLIENT_CONFIG );
    Config();
  }

  if ( ( argc == 2 ) && ( v3StrCmpA( (ascii *) "-config", (ascii *) argv[1],
    v3u64u32( 16 ) ) == 0 ) )
  {
    Config();
  }

  /* how much work do we want, do it here to minimize connection time later */
  /* calculate 24 hours of work for __speed */
  /*
  __speed = v3u64Mul( __speed, v3u64u32( 86400 ) );
  */
  /* set limit to about 15 minutes of work */
  /*
  limit = v3u64Mul( __speed, v3u64u32( 900 ) );
  */

  /* use fast numbers for example, 1 sec of work and 1/4 second limit */
  __speed = Speed();
  limit = v3u64Div( __speed, v3u64u32( 4 ) );

  while ( __shutdown_flag == 0 )
  {
    if ( ( count = GetWork( &__work_pool[0], POOL_SIZE ) ) == 0 )
    {
      v3PrintA( (ascii *) "GetWork Failed\n" );
      v3ProcessEnd( -1 );
    }

    for ( i = 0 ; i < count ; i++ )
    {
      while ( ( Run( &__work_pool[i], limit ) == STATE_ALIVE ) &&
        ( __shutdown_flag == 0 ) )
      {
        /* checkpoint */
        Save();
        v3PrintA( (ascii *) "%u Sum %016Y\n", i, __work_pool[i].total );
      }
    }

    PutWork( &__work_pool[0], POOL_SIZE );
  }

  return( V3_PASS );
}
