/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_COSM_H
#define V3_COSM_H

/* include everything we need */

#include "cputypes.h"
#include "cosmfile.h"
#include "cosmio.h"
#include "cosmmath.h"
#include "cosmmem.h"
#include "cosmnet.h"
#include "cosmtask.h"
#include "buffer.h"
#include "email.h"
#include "log.h"
#include "config.h"
#include "security.h"
#include "cosmtime.h"
#include "speed.h"
#include "compress.h"
#include "language.h"
#include "cosmgui.h"

/* testing */

#define V3_TEST_MODULES { (ascii *) "none", (ascii *) "cosmmath", \
  (ascii *) "cosmtask", (ascii *) "cosmmem", (ascii *) "cosmfile", \
  (ascii *) "cosmio", (ascii *) "cosmnet", (ascii *) "buffer", \
  (ascii *) "email", (ascii *) "log", (ascii *) "config", \
  (ascii *) "security", (ascii *) "cosmtime", (ascii *) "speed", \
  (ascii *) "compress", (ascii *) "language", (ascii *) "cosmgui", \
  (ascii *) NULL }

s32 v3Test( s32 * failed_module, s32 * failed_test, s32 module_num );
  /*
    This function is in cosmtest.c
    If module_num is 0, then test all Cosm functions and return a negative
    number corresponding to the module that failed. failed_module and
    failed_test will be set to the corresponding failed functions.
    Otherwise test only the module corresponding to module_num.
    Returns: V3_PASS on success, or a negative number corresponding to the
      module that failed.
  */

/*
  We need to define macros to catch any unsafe function use.
  Any working compiler should spit out warnings.
*/
#define gets( a ) unsafe_gets_use_v3InputA()
#define memcopy( a, b, c ) unsafe_memcpy_use_v3MemCopy()
#define strcpy( a, b ) unsafe_strcpy_use_v3StrCopyA()
#define strncpy( a, b, c ) unsafe_strcpy_use_v3StrCopyA()

/*
  We need to define macros to catch any depreciated and non-portable
  function use. Any working compiler should spit out warnings.
*/
#if ( !defined( NONPORTABLE_CODE ) )
#define abort( ) replace_abort_use_v3ProcessEnd()
#define exit( a ) replace_exit_use_v3ProcessEnd()
#define fflush( a ) replace_fflush_use_v3FileOpen()
#define fopen( a, b ) replace_fopen_use_v3FileOpen()
#define freopen( a, b, c ) replace_freopen_use_v3FileOpen()
#define printf( a ) replace_printf_use_v3PrintA()
#define raise( a ) replace_raise_use_v3ProcessSignal()
#endif /* NONPORTABLE_CODE */

#endif
