/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#include "cosmmem.h"
#include "cosmio.h"
#include <stdio.h>
#if ( OS_TYPE == OS_SUNOS )
/* Non-ANSI SunOS fails to prototype *cough*Everything*cough* */
extern int _filbuf();
extern int _flsbuf();
extern void fflush();
#endif
#if ( ( OS_TYPE == OS_WIN32 ) || ( OS_TYPE == OS_OS2 ) )
#include <io.h>
#include <conio.h>
#else
#include <unistd.h>
#if ( ( OS_TYPE == OS_LINUX ) || ( OS_TYPE == OS_FREEBSD ) \
  || ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) \
  || ( OS_TYPE == OS_SUNOS ) || ( OS_TYPE == OS_SOLARIS ) \
  || ( OS_TYPE == OS_TRU64 ) )
#include <termios.h>
#endif
#endif

/* states for the print state machine */
enum V3_IO_STATES
{
  V3_IO_TEXT,
  V3_IO_PERCENT,
  V3_IO_WIDTH,
  V3_IO_PREC,
  V3_IO_SINGLE_CHAR,
  V3_IO_STRING,
  V3_IO_BUFFER,
  V3_IO_TMP_STRING,
  V3_IO_U32_DEC,
  V3_IO_U64_DEC,
  V3_IO_U128_DEC,
  V3_IO_S32_DEC,
  V3_IO_S64_DEC,
  V3_IO_S128_DEC,
  V3_IO_U32_HEX,
  V3_IO_U64_HEX,
  V3_IO_U128_HEX,
  V3_IO_F32,
  V3_IO_F64,
  V3_IO_F32_SCI,
  V3_IO_F64_SCI,
  V3_IO_POINTER
};

#define V3_IO_NO  0
#define V3_IO_YES 1

u64 v3Input( u8 * buffer, u64 length, u32 echo, u32 wait )
{
  /*
    implementors note: ( delete after implemented )
    This is roughly equivalent to v3FileRead() only for stdio.
    Input is of course not echoed. This is basicly here for redirected input.
    Should be alot simpler, no mutex or other mess needed.
  */
}

u64 v3InputA( ascii * buffer, u64 max_chars, u32 echo )
{
  u64 chars_read;
  int tmp;
  int done;
#if ( ( OS_TYPE == OS_LINUX ) || ( OS_TYPE == OS_FREEBSD ) \
  || ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) \
  || ( OS_TYPE == OS_SUNOS ) || ( OS_TYPE == OS_SOLARIS ) \
  || ( OS_TYPE == OS_TRU64 ) )
  struct termios new_ts, old_ts;
#endif

  _V3_SET64( chars_read, 00000000, 00000000 );

  /* Make sure that buffer is actually set to something */
  if ( ( buffer == NULL ) || ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    return( chars_read );
  }

  /* dont overflow the buffer */
  v3u64Dec( &max_chars );

  /* make sure we are not buffering input */
  setvbuf( stdin, NULL, _IONBF, 0 );

  /* take care of unix echo setting */
#if ( ( OS_TYPE == OS_LINUX ) || ( OS_TYPE == OS_FREEBSD ) \
  || ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) \
  || ( OS_TYPE == OS_SUNOS ) || ( OS_TYPE == OS_SOLARIS ) \
  || ( OS_TYPE == OS_TRU64 ) )
  /* save settings */
  tcgetattr( STDIN_FILENO, &old_ts );
  new_ts = old_ts;
  if ( echo == V3_IO_NOECHO )
  {
    new_ts.c_lflag &= ~ECHO;
    new_ts.c_lflag |= ECHONL;
    tcsetattr( STDIN_FILENO, TCSANOW, &new_ts );

    /* make sure we could set them */
    tcgetattr( STDIN_FILENO, &new_ts );
    if ( ( new_ts.c_lflag & ECHO ) == ECHO )
    {
      return( chars_read );
    }
  }
  else /* echo */
  {
    new_ts.c_lflag |= ECHO;
    new_ts.c_lflag |= ECHONL;
    tcsetattr( STDIN_FILENO, TCSANOW, &new_ts );

    /* make sure we could set them */
    tcgetattr( STDIN_FILENO, &new_ts );
    if ( ( new_ts.c_lflag & ECHO ) != ECHO )
    {
      return( chars_read );
    }
  }
#endif

  done = V3_IO_NO;
  while ( done == V3_IO_NO )
  {
#if ( OS_TYPE == OS_WIN32 )
    /* in Win32 and friends, getch() doesn't do echo */
    if ( echo == V3_IO_NOECHO )
    {
      if ( ( tmp = (int) getch() ) == 13 )
      {
        /* from the \n -> \r\n we need to get another char */
        tmp = fgetc( stdin );
      }
    }
    else
    {
      tmp = fgetc( stdin );
    }
#else
    tmp = fgetc( stdin );
#endif

    /* read until the linefeed, place chars in buffer if under limit */
    switch ( tmp )
    {
      case -1: /* EOF */
        if ( v3u64Eq( chars_read, v3u64u32( 0 ) ) )
        {
          /* no characters read yet, then EOF */
          _V3_SET64( chars_read, FFFFFFFF, FFFFFFFF );
        }
        done = V3_IO_YES;
        break;
      case '\n': /* End of Line */
        done = V3_IO_YES;
        break;
      default:
        if ( v3u64Lt( chars_read, max_chars ) )
        {
          *buffer++ = (ascii) tmp;
          v3u64Inc( &chars_read );
        }
        break;
    }
  }

#if ( ( OS_TYPE == OS_LINUX ) || ( OS_TYPE == OS_FREEBSD ) \
  || ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) \
  || ( OS_TYPE == OS_SUNOS ) || ( OS_TYPE == OS_SOLARIS ) \
  || ( OS_TYPE == OS_TRU64 ) )
  /* restore echo state and flush here */
  tcsetattr( STDIN_FILENO, TCSAFLUSH, &old_ts );
#endif

  /* terminate string and leave */
  *buffer = 0;
  return( chars_read );
}

u64 v3InputU( unicode * buffer, u64 max_chars, u32 echo )
{
  /* !!! INSERT CODE HERE */
  u64 ret;
  _V3_SET64( ret, FFFFFFFF, FFFFFFFF );
  return( ret );
}

u64 v3StrLengthA( const ascii * string )
{
  u64 length;

  _V3_SET64( length, 00000000, 00000000 );

  if ( string == NULL )
  {
    return( length );
  }

  while ( *(string)++ != 0 )
  {
    v3u64Inc( &length );
  }

  return( length );
}

u64 v3StrLengthU( const unicode * string )
{
  u64 length;

  _V3_SET64( length, 00000000, 00000000 );

  if ( string == NULL )
  {
    return( length );
  }

  while ( *(string)++ != 0 )
  {
    v3u64Inc( &length );
  }

  return( length );
}

s32 v3StrCopyA( ascii * dest, const ascii * src, u64 max_chars )
{
  u64 length;

  if ( ( dest == NULL ) || ( src == NULL ) ||
    ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    return( V3_FAIL );
  }

  /* check for ( max_chars - 1 ) */
  length = v3StrLengthA( src );
  v3u64Dec( &max_chars );

  if ( v3u64Gt( length, max_chars ) )
  {
    return( V3_FAIL );
  }

  while ( v3u64Gt( length, v3u64u32( 0 ) ) )
  {
    *(dest)++ = *(src)++;
    v3u64Dec( &length );
  }
  *dest = 0;

  return( V3_PASS );
}

s32 v3StrCopyU( unicode * dest, const unicode * src, u64 max_chars )
{
  u64 length;

  if ( ( dest == NULL ) || ( src == NULL ) ||
    ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    return( V3_FAIL );
  }

  /* check for ( max_chars - 1 ) */
  length = v3StrLengthU( src );
  v3u64Dec( &max_chars );

  if ( v3u64Gt( length, max_chars ) )
  {
    return( V3_FAIL );
  }

  while ( v3u64Gt( length, v3u64u32( 0 ) ) )
  {
    *(dest)++ = *(src)++;
    v3u64Dec( &length );
  }
  *dest = 0;

  return( V3_PASS );
}

s32 v3StrCopyUA( unicode * dest, const ascii * src, u64 max_chars )
{
  u64 length;

  if ( ( dest == NULL ) || ( src == NULL ) ||
    ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    return( V3_FAIL );
  }

  /* check for ( max_chars - 1 ) */
  length = v3StrLengthA( src );
  v3u64Dec( &max_chars );

  if ( v3u64Gt( length, max_chars ) )
  {
    return( V3_FAIL );
  }

  /* Use the counter for the length, but use pointers for addressing */

  while ( v3u64Gt( length, v3u64u32( 0 ) ) )
  {
    *(dest)++ = (unicode) *(src)++;
    v3u64Dec( &length );
  }
  *dest = (unicode) 0;

  return( V3_PASS );
}

s32 v3StrCmpA( const ascii * stringA, const ascii * stringB, u64 max_chars )
{
  u64 i;
  s32 a, b;

  if ( ( stringA == NULL ) || ( stringB == NULL )
    || ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    if ( stringA != stringB )
    {
      return( -1 );
    }
    return( 0 );
  }

  _V3_SET64( i, 00000000, 00000000 );
  do
  {
    a = (s32) *(stringA)++;
    b = (s32) *(stringB)++;
    v3u64Inc( &i );
  } while ( ( a == b ) && ( a != 0 ) && v3u64Lt( i, max_chars ) );

  return( a - b );
}

s32 v3StrCmpU( const unicode * stringA, const unicode * stringB,
  u64 max_chars )
{
  u64 i;
  s32 a, b;

  if ( ( stringA == NULL ) || ( stringB == NULL )
    || ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    if ( stringA != stringB )
    {
      return( -1 );
    }
    return( 0 );
  }

  _V3_SET64( i, 00000000, 00000000 );
  do
  {
    a = (s32) *(stringA)++;
    b = (s32) *(stringB)++;
    v3u64Inc( &i );
  } while ( ( a == b ) && ( a != 0 ) && v3u64Lt( i, max_chars ) );

  return( a - b );
}

ascii * v3StrCharA( const ascii * string, ascii match_char, u64 max_chars )
{
  u64 count;

  if ( ( string == NULL ) || ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    return( (ascii *) NULL );
  }

  _V3_SET64( count, 00000000, 00000000 );

  while ( ( v3u64Lt( count, max_chars ) ) && ( *string != 0 ) )
  {
    if ( *string == match_char )
    {
      return( (ascii *) string );
    }
    v3u64Inc( &count );
    string++;
  }

  return( (ascii *) NULL );
}

unicode * v3StrCharU( const unicode * string, unicode match_char,
  u64 max_chars )
{
  u64 count;

  if ( ( string == NULL ) || ( v3u64Eq( max_chars, v3u64u32( 0 ) ) ) )
  {
    return( (unicode *) NULL );
  }

  _V3_SET64( count, 00000000, 00000000 );

  while ( ( v3u64Lt( count, max_chars ) ) && ( *string != 0 ) )
  {
    if ( *string == match_char )
    {
      return( (unicode *) string );
    }
    v3u64Inc( &count );
    string++;
  }

  return( (unicode *) NULL );
}

ascii * v3StrStrA( const ascii * string, const ascii * substring,
  u64 max_chars )
{
  ascii * first, * end, * str, * sub;
  ascii ch;
  u64 len;

  /* check for bad or trivial parameters */
  if ( ( string == NULL ) || ( substring == NULL ) ||
    ( string[0] == 0 ) || ( substring[0] == 0 ) )
  {
    return( (ascii *) NULL );
  }

  /* substring too long? */
  if ( v3u64Gt( len = v3StrLengthA( substring ),
    v3u64Sub( max_chars, v3u64u32( 1 ) ) ) )
  {
    return( (ascii *) NULL );
  }

  /*
    end is set to the last character substring could safely start at.
    makes it safe and saves some iterations too.
  */
  v3u64Inc( &len );
  end = (ascii *) v3MemOffset( string, v3u64Sub( max_chars, len ) );

  first = (ascii *) string;
  ch = substring[0];

  for ( ; ; )
  {
    /* find a matching first character in string */
    while ( *first != ch )
    {
      if ( *first == 0 )
      {
        return( (ascii *) NULL );
      }
      first++;
      if ( first > end )
      {
        return( (ascii *) NULL );
      }
    }

    str = first;
    sub = (ascii *) substring;

    /* match up the strings */
    while ( ( *str == *sub ) && ( *sub != 0 ) )
    {
      str++;
      sub++;
    }

    /* if we got safely to the end of sub, then we found it */
    if ( *sub == 0 )
    {
      /* check for really bad case - string was unterminated */
      if ( ( first == end ) && ( *str != 0 ) )
      {
        return( (ascii *) NULL );
      }

      return( first );
    }

    first++;
    if ( first > end )
    {
      return( (ascii *) NULL );
    }
  }
}

unicode * v3StrStrU( const unicode * string, const unicode * substring,
  u64 max_chars )
{
  unicode * first, * end, * str, * sub;
  unicode ch;
  u64 len;

  /* check for bad or trivial parameters */
  if ( ( string == NULL ) || ( substring == NULL ) ||
    ( string[0] == 0 ) || ( substring[0] == 0 ) )
  {
    return( (unicode *) NULL );
  }

  /* substring too long? */
  if ( v3u64Gt( len = v3StrLengthU( substring ),
    v3u64Sub( max_chars, v3u64u32( 1 ) ) ) )
  {
    return( (unicode *) NULL );
  }

  /*
    end is set to the last character substring could safely start at.
    makes it safe and saves some iterations too.
  */
  v3u64Inc( &len );
  end = (unicode *) v3MemOffset( string,
    v3u64Lsh( v3u64Sub( max_chars, len ), 1 ) );

  first = (unicode *) string;
  ch = substring[0];

  for ( ; ; )
  {
    /* find a matching first character in string */
    while ( *first != ch )
    {
      if ( *first == 0 )
      {
        return( (unicode *) NULL );
      }
      first++;
      if ( first > end )
      {
        return( (unicode *) NULL );
      }
    }

    str = first;
    sub = (unicode *) substring;

    /* match up the strings */
    while ( ( *str == *sub ) && ( *sub != 0 ) )
    {
      str++;
      sub++;
    }

    /* if we got safely to the end of sub, then we found it */
    if ( *sub == 0 )
    {
      /* check for really bad case - string was unterminated */
      if ( ( first == end ) && ( *str != 0 ) )
      {
        return( (unicode *) NULL );
      }

      return( first );
    }

    first++;
    if ( first > end )
    {
      return( (unicode *) NULL );
    }
  }
}

s32 v3f32A( f32 * result, ascii ** end, const ascii * string )
{

}

s32 v3f64A( f64 * result, ascii ** end, const ascii * string )
{

}

s32 v3f32U( f32 * result, unicode ** end, const unicode * string )
{

}

s32 v3f64U( f64 * result, unicode ** end, const unicode * string )
{

}

u64 v3PrintA( const ascii * format, ... )
{
  u64 chars_output;
  u64 tmp;
  va_list ap;

  if ( format == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  _V3_SET64( tmp, FFFFFFFF, FFFFFFFF );

  va_start( ap, format );
  chars_output = v3_Print( (v3_FILE *) NULL, NULL, 1, tmp, format, ap );
  va_end( ap );

  fflush( stdout );

  return( chars_output );
}

u64 v3PrintU( const unicode * format, ... )
{
  u64 chars_output;
  u64 tmp;
  va_list ap;

  if ( format == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  _V3_SET64( tmp, FFFFFFFF, FFFFFFFF );

  va_start( ap, format );
  chars_output = v3_Print( (v3_FILE *) NULL, NULL, 2, tmp, format, ap );
  va_end( ap );

  fflush( NULL );

  return( chars_output );
}

u64 v3PrintAStr( ascii * string, u64 max_chars, const ascii * format, ... )
{
  u64 chars_output;
  va_list ap;

  if ( string == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  if ( v3u64Gt( max_chars, v3u64u32( 0 ) ) )
  {
    v3u64Dec( &max_chars );
  }

  va_start( ap, format );
  chars_output = v3_Print( (v3_FILE *) NULL, string, 1, max_chars, format,
    ap );
  va_end( ap );

  *( (ascii * ) v3MemOffset( string, chars_output ) ) = 0;

  return( chars_output );
}

u64 v3PrintUStr( unicode * string, u64 max_chars,
                 const unicode * format, ... )
{
  u64 chars_output;
  va_list ap;

  if ( string == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  if ( v3u64Gt( max_chars, v3u64u32( 0 ) ) )
  {
    v3u64Dec( &max_chars );
  }

  va_start( ap, format );
  chars_output = v3_Print( (v3_FILE *) NULL, string, 2, max_chars, format,
    ap );
  va_end( ap );

  *( (unicode * ) v3MemOffset( string, v3u64Lsh( chars_output, 1 ) ) ) = 0;

  return( chars_output );
}

u64 v3PrintAFile( v3_FILE * file, const ascii * format, ... )
{
  u64 chars_output;
  u64 tmp;
  va_list ap;

  _V3_SET64( tmp, FFFFFFFF, FFFFFFFF );

  if ( file == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  va_start( ap, format );
  chars_output = v3_Print( file, NULL, 1, tmp, format, ap );
  va_end( ap );

  return( chars_output );
}

u64 v3PrintUFile( v3_FILE * file, const unicode * format, ... )
{
  u64 chars_output;
  u64 tmp;
  va_list ap;

  _V3_SET64( tmp, FFFFFFFF, FFFFFFFF );

  if ( file == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  va_start( ap, format );
  chars_output = v3_Print( file, NULL, 2, tmp, format, ap );
  va_end( ap );

  return( chars_output );
}

/* Low level functions */

s32 v3_ParseTextA( void * number, u8 number_width, u32 sign, ascii ** end,
  const ascii * string, u32 radix )
{
  u32  tmp_u32, max_u32s, max_rem;
  u64  tmp_u64, max_u64s, rad_u64;
  u128 tmp_u128, max_u128s, rad_u128;
  ascii ch;
  u32 minus;
  s8 val;
  static const s8 num[256] =
  {
    -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -2, -1, -2, -2,
     0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -2, -2, -2, -2, -2, -2,
    -2, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
    25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, -2, -2, -2, -2, -2,
    -2, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
    25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
    -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2
  };

  if ( ( number == NULL ) || ( string == NULL ) || ( radix > 36 ) ||
    ( radix == 1 ) )
  {
    return( V3_FAIL );
  }

  ch = *string;
  val = num[ch];

  /* skip whitespaces (val -3 stuff) */
  while ( val == -3 )
  {
    ch = *(++string);
    val = num[ch];
  }

  /* sign? (val -1) */
  minus = 0;
  if ( val == -1 )
  {
    if ( ch == '-' )
    {
      if ( sign == V3_IO_SIGNED )
      {
        minus = 1;
      }
      else
      {
        return( V3_FAIL );
      }
    }
    ch = *(++string);
  }

  /* determine radix if we need to */
  if ( radix == 0 )
  {
    if ( ch == '0' )
    {
      radix = 8;
      ch = *(++string);
      if ( ( ch == 'x' ) || ( ch == 'X' ) )
      {
        radix = 16;
        ch = *(++string);
      }
    }
    else
    {
      radix = 10;
    }
  }
  /* resync val */
  val = num[ch];

  /*
    parse the numbers until we hit something with val above radix or < 0
    which should happen before we overflow the number.
  */

  /*
    setup limits, and start eatting numbers
    max_...s is the max a number can be and still shift.
    max_rem is the max final remainder we can have at that last shift.
  */
  switch ( number_width )
  {
    case 4:
      tmp_u32 = 0;
      max_u32s = 0xFFFFFFFF;
      if ( sign == V3_IO_SIGNED )
      {
        max_u32s >>= 1;
      }
      max_rem = max_u32s % radix;
      max_u32s = max_u32s / radix;
      /* eat numbers */
      while ( ( val >= 0 ) && ( val < (s32) radix ) )
      {
        if ( ( tmp_u32 >= max_u32s ) ||
          ( ( tmp_u32 == max_u32s ) && ( val > (s32) max_rem ) ) )
        {
          /* too many characters */
          return( V3_FAIL );
        }
        tmp_u32 *= radix;
        tmp_u32 += val;
        ch = *(++string);
        val = num[ch];
      }
      /* fix sign, set number */
      if ( minus == 1 )
      {
        tmp_u32 = ~tmp_u32 + 1;
      }
      *( (u32 *) number ) = tmp_u32;
      break;
    case 8:
      _V3_SET64( tmp_u64, 00000000, 00000000 );
      _V3_SET64( max_u64s, FFFFFFFF, FFFFFFFF );
      if ( sign == V3_IO_SIGNED )
      {
        max_u64s = v3u64Rsh( max_u64s, 1 );
      }
      rad_u64 = v3u64u32( radix );
      max_rem = v3u32u64( v3u64Mod( max_u64s, rad_u64 ) );
      max_u64s = v3u64Div( max_u64s, rad_u64 );
      /* eat numbers */
      while ( ( val >= 0 ) && ( val < (s32) radix ) )
      {
        if ( ( !v3u64Lt( tmp_u64, max_u64s ) ) ||
          ( ( v3u64Eq( tmp_u64, max_u64s ) ) && ( val > (s32) max_rem ) ) )
        {
          /* too many characters */
          return( V3_FAIL );
        }
        tmp_u64 = v3u64Add( v3u64Mul( tmp_u64, rad_u64 ), v3u64u32( val ) );
        ch = *(++string);
        val = num[ch];
      }
      /* fix sign, set number */
      if ( minus == 1 )
      {
        tmp_u64 = v3u64Add( v3u64Not( tmp_u64 ), v3u64u32( 1 ) );
      }
      *( (u64 *) number ) = tmp_u64;
      break;
    case 16:
      _V3_SET128( tmp_u128, 00000000, 00000000, 00000000, 00000000 );
      _V3_SET128( max_u128s, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF );
      if ( sign == V3_IO_SIGNED )
      {
        max_u128s = v3u128Rsh( max_u128s, 1 );
      }
      rad_u128 = v3u128u32( radix );
      max_rem = v3u32u128( v3u128Mod( max_u128s, rad_u128 ) );
      max_u128s = v3u128Div( max_u128s, rad_u128 );
      /* eat numbers */
      while ( ( val >= 0 ) && ( val < (s32) radix ) )
      {
        if ( ( !v3u128Lt( tmp_u128, max_u128s ) ) ||
          ( ( v3u128Eq( tmp_u128, max_u128s ) ) && ( val > (s32) max_rem ) ) )
        {
          /* too many characters */
          return( V3_FAIL );
        }
        tmp_u128 = v3u128Add( v3u128Mul( tmp_u128, rad_u128 ),
          v3u128u32( val ) );
        ch = *(++string);
        val = num[ch];
      }
      /* fix sign, set number */
      if ( minus == 1 )
      {
        tmp_u128 = v3u128Add( v3u128Not( tmp_u128 ), v3u128u32( 1 ) );
      }
      *( (u128 *) number ) = tmp_u128;
      break;
    default:
      return( V3_FAIL );
  }

  /* set end if we need to */
  if ( end != NULL )
  {
    *end = (ascii *) string;
  }

  return( V3_PASS );
}

s32 v3_ParseTextU( void * number, u8 number_width, u32 sign, unicode ** end,
  const unicode * string, u32 radix )
{
  u32  tmp_u32, max_u32s, max_rem;
  u64  tmp_u64, max_u64s, rad_u64;
  u128 tmp_u128, max_u128s, rad_u128;
  unicode ch;
  u32 minus;
  s8 val;

  if ( ( number == NULL ) || ( string == NULL ) || ( radix > 36 ) ||
    ( radix == 1 ) )
  {
    return( V3_FAIL );
  }

  ch = *string;
  val = v3_GetCharValU( ch );

  /* skip whitespaces (val -3 stuff) */
  while ( val == -3 )
  {
    ch = *(++string);
    val = v3_GetCharValU( ch );
  }

  /* sign? (val -1) */
  minus = 0;
  if ( val == -1 )
  {
    if ( ch == '-' )
    {
      if ( sign == V3_IO_SIGNED )
      {
        minus = 1;
      }
      else
      {
        return( V3_FAIL );
      }
    }
    ch = *(++string);
  }

  /* determine radix if we need to */
  if ( radix == 0 )
  {
    if ( ch == '0' )
    {
      radix = 8;
      ch = *(++string);
      if ( ( ch == 'x' ) || ( ch == 'X' ) )
      {
        radix = 16;
        ch = *(++string);
      }
    }
    else
    {
      radix = 10;
    }
  }
  /* resync val */
  val = v3_GetCharValU( ch );

  /*
    parse the numbers until we hit something with val above radix or < 0
    which should happen before we overflow the number.
  */

  /*
    setup limits, and start eatting numbers
    max_...s is the max a number can be and still shift.
    max_rem is the max final remainder we can have at that last shift.
  */
  switch ( number_width )
  {
    case 4:
      tmp_u32 = 0;
      max_u32s = 0xFFFFFFFF;
      if ( sign == V3_IO_SIGNED )
      {
        max_u32s >>= 1;
      }
      max_rem = max_u32s % radix;
      max_u32s = max_u32s / radix;
      /* eat numbers */
      while ( ( val >= 0 ) && ( val < (s32) radix ) )
      {
        if ( ( tmp_u32 >= max_u32s ) ||
          ( ( tmp_u32 == max_u32s ) && ( val > (s32) max_rem ) ) )
        {
          /* too many characters */
          return( V3_FAIL );
        }
        tmp_u32 *= radix;
        tmp_u32 += val;
        ch = *(++string);
        val = v3_GetCharValU( ch );
      }
      /* fix sign, set number */
      if ( minus == 1 )
      {
        tmp_u32 = ~tmp_u32 + 1;
      }
      *( (u32 *) number ) = tmp_u32;
      break;
    case 8:
      _V3_SET64( tmp_u64, 00000000, 00000000 );
      _V3_SET64( max_u64s, FFFFFFFF, FFFFFFFF );
      if ( sign == V3_IO_SIGNED )
      {
        max_u64s = v3u64Rsh( max_u64s, 1 );
      }
      rad_u64 = v3u64u32( radix );
      max_rem = v3u32u64( v3u64Mod( max_u64s, rad_u64 ) );
      max_u64s = v3u64Div( max_u64s, rad_u64 );
      /* eat numbers */
      while ( ( val >= 0 ) && ( val < (s32) radix ) )
      {
        if ( ( !v3u64Lt( tmp_u64, max_u64s ) ) ||
          ( ( v3u64Eq( tmp_u64, max_u64s ) ) && ( val > (s32) max_rem ) ) )
        {
          /* too many characters */
          return( V3_FAIL );
        }
        tmp_u64 = v3u64Add( v3u64Mul( tmp_u64, rad_u64 ), v3u64u32( val ) );
        ch = *(++string);
        val = v3_GetCharValU( ch );
      }
      /* fix sign, set number */
      if ( minus == 1 )
      {
        tmp_u64 = v3u64Add( v3u64Not( tmp_u64 ), v3u64u32( 1 ) );
      }
      *( (u64 *) number ) = tmp_u64;
      break;
    case 16:
      _V3_SET128( tmp_u128, 00000000, 00000000, 00000000, 00000000 );
      _V3_SET128( max_u128s, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF );
      if ( sign == V3_IO_SIGNED )
      {
        max_u128s = v3u128Rsh( max_u128s, 1 );
      }
      rad_u128 = v3u128u32( radix );
      max_rem = v3u32u128( v3u128Mod( max_u128s, rad_u128 ) );
      max_u128s = v3u128Div( max_u128s, rad_u128 );
      /* eat numbers */
      while ( ( val >= 0 ) && ( val < (s32) radix ) )
      {
        if ( ( !v3u128Lt( tmp_u128, max_u128s ) ) ||
          ( ( v3u128Eq( tmp_u128, max_u128s ) ) && ( val > (s32) max_rem ) ) )
        {
          /* too many characters */
          return( V3_FAIL );
        }
        tmp_u128 = v3u128Add( v3u128Mul( tmp_u128, rad_u128 ),
          v3u128u32( val ) );
        ch = *(++string);
        val = v3_GetCharValU( ch );
      }
      /* fix sign, set number */
      if ( minus == 1 )
      {
        tmp_u128 = v3u128Add( v3u128Not( tmp_u128 ), v3u128u32( 1 ) );
      }
      *( (u128 *) number ) = tmp_u128;
      break;
    default:
      return( V3_FAIL );
  }

  /* set end if we need to */
  if ( end != NULL )
  {
    *end = (unicode *) string;
  }

  return( V3_PASS );
}

s8 v3_GetCharValU( unicode ch )
{
  s8 val;

  /*
    AUTO GENERATED FROM
    ftp://ftp.unicode.org/Public/UNIDATA/UnicodeData-Latest.txt on 7/15/00
    Then hand edited for sanity. Then added to by hand because the layout
    of the symbols in each section is not consistant --HAA
  */

  /* whitespace */
  if ( ( ch >= 0x0008 && ch <= 0x000D ) || /* returns, etc.. */
         ch == 0x0020 || /* SPACE */
         ch == 0x00A0 || /* NO-BREAK SPACE */
         ch == 0x1361 || /* ETHIOPIC WORDSPACE */
         ch == 0x1680 || /* OGHAM SPACE MARK */
       ( ch >= 0x2002 && ch <= 0x200B ) || /* different size spaces */
         ch == 0x202F || /* NARROW NO-BREAK SPACE */
         ch == 0x3000 || /* IDEOGRAPHIC SPACE */
         ch == 0x303F || /* IDEOGRAPHIC HALF FILL SPACE */
         ch == 0xFEFF ) /* ZERO WIDTH NO-BREAK SPACE */
  {
   val = -3;
  }

  /* plain letters, ie without accents */
  /* LATIN CAPITAL LETTER A through LATIN CAPITAL LETTER Z */
  else if ( ch >= 0x0041 && ch <= 0x005A )
  {
    val = (s8) ( ( ch - 0x0041 ) + 10 );
  }
  /* LATIN SMALL LETTER A through LATIN SMALL LETTER Z */
  else if ( ch >= 0x0061 && ch <= 0x007A )
  {
    val = (s8) ( ( ch - 0x0061 ) + 10 );
  }
  /* GREEK CAPITAL LETTER ALPHA through GREEK CAPITAL LETTER UPSILON */
  else if ( ch >= 0x0391 && ch <= 0x03A5 )
  {
    val = (s8) ( ( ch - 0x0391 ) + 10 );
  }
  /* GREEK SMALL LETTER ALPHA through GREEK SMALL LETTER UPSILON */
  else if ( ch >= 0x03AC && ch <= 0x03C5 )
  {
    val = (s8) ( ( ch - 0x03AC ) + 10 );
  }
  /* ARMENIAN CAPITAL LETTER AYB through ARMENIAN CAPITAL LETTER FEH */
  else if ( ch >= 0x0531 && ch <= 0x0556 )
  {
    val = (s8) ( ( ch - 0x0531 ) + 10 );
  }
  /* ARMENIAN SMALL LETTER AYB through ARMENIAN SMALL LETTER FEH */
  else if ( ch >= 0x0561 && ch <= 0x0586 )
  {
    val = (s8) ( ( ch - 0x0561 ) + 10 );
  }
  /* HEBREW LETTER ALEF through HEBREW LETTER TAV */
  else if ( ch >= 0x05D0 && ch <= 0x05EA )
  {
    val = (s8) ( ( ch - 0x05D0 ) + 10 );
  }
  /* ARABIC LETTER ALEF through ARABIC LETTER YEH */
  else if ( ch >= 0x0627 && ch <= 0x064A )
  {
    val = (s8) ( ( ch - 0x0627 ) + 10 );
  }
  /* SYRIAC LETTER ALAPH through SYRIAC LETTER TAW */
  else if ( ch >= 0x0710 && ch <= 0x072C )
  {
    val = (s8) ( ( ch - 0x0710 ) + 10 );
  }
  /* THAANA LETTER HAA through THAANA LETTER WAAVU */
  else if ( ch >= 0x0780 && ch <= 0x07A5 )
  {
    val = (s8) ( ( ch - 0x0780 ) + 10 );
  }
  /* DEVANAGARI LETTER A through DEVANAGARI LETTER HA */
  else if ( ch >= 0x0905 && ch <= 0x0939 )
  {
    val = (s8) ( ( ch - 0x0905 ) + 10 );
  }
  /* BENGALI LETTER A through BENGALI LETTER HA */
  else if ( ch >= 0x0985 && ch <= 0x09B9 )
  {
    val = (s8) ( ( ch - 0x0985 ) + 10 );
  }
  /* GURMUKHI LETTER A through GURMUKHI LETTER HA */
  else if ( ch >= 0x0A05 && ch <= 0x0A39 )
  {
    val = (s8) ( ( ch - 0x0A05 ) + 10 );
  }
  /* GUJARATI LETTER A through GUJARATI LETTER HA */
  else if ( ch >= 0x0A85 && ch <= 0x0AB9 )
  {
    val = (s8) ( ( ch - 0x0A85 ) + 10 );
  }
  /* ORIYA LETTER A through ORIYA LETTER HA */
  else if ( ch >= 0x0B05 && ch <= 0x0B39 )
  {
    val = (s8) ( ( ch - 0x0B05 ) + 10 );
  }
  /* TAMIL LETTER A through TAMIL LETTER HA */
  else if ( ch >= 0x0B85 && ch <= 0x0BB9 )
  {
    val = (s8) ( ( ch - 0x0B85 ) + 10 );
  }
  /* TELUGU LETTER A through TELUGU LETTER HA */
  else if ( ch >= 0x0C05 && ch <= 0x0C39 )
  {
    val = (s8) ( ( ch - 0x0C05 ) + 10 );
  }
  /* KANNADA LETTER A through KANNADA LETTER HA */
  else if ( ch >= 0x0C85 && ch <= 0x0CB9 )
  {
    val = (s8) ( ( ch - 0x0C85 ) + 10 );
  }
  /* MALAYALAM LETTER A through MALAYALAM LETTER HA */
  else if ( ch >= 0x0D05 && ch <= 0x0D39 )
  {
    val = (s8) ( ( ch - 0x0D05 ) + 10 );
  }
  /* SINHALA LETTER AYANNA through SINHALA LETTER FAYANNA */
  else if ( ch >= 0x0D85 && ch <= 0x0DC6 )
  {
    val = (s8) ( ( ch - 0x0D85 ) + 10 );
  }
  /* THAI CHARACTER KO KAI through THAI CHARACTER HO NOKHUK */
  else if ( ch >= 0x0E01 && ch <= 0x0E2E )
  {
    val = (s8) ( ( ch - 0x0E01 ) + 10 );
  }
  /* LAO LETTER KO through LAO LETTER HO TAM */
  else if ( ch >= 0x0E81 && ch <= 0x0EAE )
  {
    val = (s8) ( ( ch - 0x0E81 ) + 10 );
  }
  /* TIBETAN LETTER KA through TIBETAN LETTER KSSA */
  else if ( ch >= 0x0F40 && ch <= 0x0F69 )
  {
    val = (s8) ( ( ch - 0x0F40 ) + 10 );
  }
  /* MYANMAR LETTER KA through MYANMAR LETTER AU */
  else if ( ch >= 0x1000 && ch <= 0x102A )
  {
    val = (s8) ( ( ch - 0x1000 ) + 10 );
  }
  /* GEORGIAN CAPITAL LETTER AN through GEORGIAN CAPITAL LETTER HOE */
  else if ( ch >= 0x10A0 && ch <= 0x10C5 )
  {
    val = (s8) ( ( ch - 0x10A0 ) + 10 );
  }
  /* GEORGIAN LETTER AN through GEORGIAN LETTER HOE */
  else if ( ch >= 0x10D0 && ch <= 0x10F5 )
  {
    val = (s8) ( ( ch - 0x10D0 ) + 10 );
  }
  /* CHEROKEE LETTER A through CHEROKEE LETTER YV */
  else if ( ch >= 0x13A0 && ch <= 0x13F4 )
  {
    val = (s8) ( ( ch - 0x13A0 ) + 10 );
  }
  /* OGHAM LETTER BEITH through OGHAM LETTER PEITH */
  else if ( ch >= 0x1681 && ch <= 0x169A )
  {
    val = (s8) ( ( ch - 0x1681 ) + 10 );
  }
  /* RUNIC LETTER FEHU FEOH FE F through RUNIC LETTER X */
  else if ( ch >= 0x16A0 && ch <= 0x16EA )
  {
    val = (s8) ( ( ch - 0x16A0 ) + 10 );
  }
  /* KHMER LETTER KA through KHMER LETTER QA */
  else if ( ch >= 0x1780 && ch <= 0x17A2 )
  {
    val = (s8) ( ( ch - 0x1780 ) + 10 );
  }
  /* MONGOLIAN LETTER A through MONGOLIAN LETTER CHI */
  else if ( ch >= 0x1820 && ch <= 0x1842 )
  {
    val = (s8) ( ( ch - 0x1820 ) + 10 );
  }


  /* normal digits */
  /* DIGIT ZERO through DIGIT NINE */
  else if ( ch >= 0x0030 && ch <= 0x0039 )
  {
    val = (s8) ( ch - 0x0030 );
  }
  /* ARABIC-INDIC DIGIT ZERO through ARABIC-INDIC DIGIT NINE */
  else if ( ch >= 0x0660 && ch <= 0x0669 )
  {
    val = (s8) ( ch - 0x0660 );
  }
  /*
    EXTENDED ARABIC-INDIC DIGIT ZERO through EXTENDED ARABIC-INDIC
    DIGIT NINE
  */
  else if ( ch >= 0x06F0 && ch <= 0x06F9 )
  {
    val = (s8) ( ch - 0x06F0 );
  }
  /* DEVANAGARI DIGIT ZERO through DEVANAGARI DIGIT NINE */
  else if ( ch >= 0x0966 && ch <= 0x096F )
  {
    val = (s8) ( ch - 0x0966 );
  }
  /* BENGALI DIGIT ZERO through BENGALI DIGIT NINE */
  else if ( ch >= 0x09E6 && ch <= 0x09EF )
  {
    val = (s8) ( ch - 0x09E6 );
  }
  /* GURMUKHI DIGIT ZERO through GURMUKHI DIGIT NINE */
  else if ( ch >= 0x0A66 && ch <= 0x0A6F )
  {
    val = (s8) ( ch - 0x0A66 );
  }
  /* GUJARATI DIGIT ZERO through GUJARATI DIGIT NINE */
  else if ( ch >= 0x0AE6 && ch <= 0x0AEF )
  {
    val = (s8) ( ch - 0x0AE6 );
  }
  /* ORIYA DIGIT ZERO through ORIYA DIGIT NINE */
  else if ( ch >= 0x0B66 && ch <= 0x0B6F )
  {
    val = (s8) ( ch - 0x0B66 );
  }
  /* TAMIL DIGIT ONE through TAMIL DIGIT NINE -- NOTE: No "ZERO" Digit. */
  else if ( ch >= 0x0BE7 && ch <= 0x0BEF )
  {
    val = (s8) ( ch - 0x0BE6 );
  }
  /* TELUGU DIGIT ZERO through TELUGU DIGIT NINE */
  else if ( ch >= 0x0C66 && ch <= 0x0C6F )
  {
    val = (s8) ( ch - 0x0C66 );
  }
  /* KANNADA DIGIT ZERO through KANNADA DIGIT NINE */
  else if ( ch >= 0x0CE6 && ch <= 0x0CEF )
  {
    val = (s8) ( ch - 0x0CE6 );
  }
  /* MALAYALAM DIGIT ZERO through MALAYALAM DIGIT NINE */
  else if ( ch >= 0x0D66 && ch <= 0x0D6F )
  {
    val = (s8) ( ch - 0x0D66 );
  }
  /* THAI DIGIT ZERO through THAI DIGIT NINE */
  else if ( ch >= 0x0E50 && ch <= 0x0E59 )
  {
    val = (s8) ( ch - 0x0E50 );
  }
  /* LAO DIGIT ZERO through LAO DIGIT NINE */
  else if ( ch >= 0x0ED0 && ch <= 0x0ED9 )
  {
    val = (s8) ( ch - 0x0ED0 );
  }
  /* TIBETAN DIGIT ZERO through TIBETAN DIGIT NINE */
  else if ( ch >= 0x0F20 && ch <= 0x0F29 )
  {
    val = (s8) ( ch - 0x0F20 );
  }
  /* MYANMAR DIGIT ZERO through MYANMAR DIGIT NINE */
  else if ( ch >= 0x1040 && ch <= 0x1049 )
  {
    val = (s8) ( ch - 0x1040 );
  }
  /* ETHIOPIC DIGIT ONE through ETHIOPIC DIGIT NINE -- NOTE: No "ZERO" Digit. */
  else if ( ch >= 0x1369 && ch <= 0x1371 )
  {
    val = (s8) ( ch - 0x1368 );
  }
  /* KHMER DIGIT ZERO through KHMER DIGIT NINE */
  else if ( ch >= 0x17E0 && ch <= 0x17E9 )
  {
    val = (s8) ( ch - 0x17E0 );
  }
  /* MONGOLIAN DIGIT ZERO through MONGOLIAN DIGIT NINE */
  else if ( ch >= 0x1810 && ch <= 0x1819 )
  {
    val = (s8) ( ch - 0x1810 );
  }
  else
  {
    val = -2;
  }

  return val;
}

/* The following 3 macros are used by v3_Print */

/* Macro for getting next character */
#define _V3_IO_FETCHCHAR { if ( char_width == 1 ) \
  character = *(f_ascii)++; \
else \
  character = *(f_unicode)++; }

/* Macro to reset tmp variables */
#define _V3_IO_CLEARTMP { state = V3_IO_TEXT; \
  width = 0; \
  width_set = V3_IO_NO; \
  pad_left = V3_IO_NO; \
  pad_zero = V3_IO_NO; \
  precision = 6; \
  precision_set = V3_IO_NO; \
  tmp_count = 0; \
  tmp_sign = 0; \
  ptr_ascii = &tmp_ascii[0]; }

/* Macro for makeing sure we don't print too much */
#define _V3_IO_CHECKMAX { v3u64Inc( &chars_output ); \
  if ( v3u64Eq( chars_output, max_chars ) ) \
    return( chars_output ); \
  if ( string != NULL ) \
  { \
    if ( char_width == 1 ) \
    { \
      s_ascii++; \
      string = s_ascii;  \
    } \
    else \
    { \
      s_unicode++; \
      string = s_unicode; \
    } \
  } }

u64 v3_Print( v3_FILE * file, void * string, u8 char_width, u64 max_chars,
  const void * format, va_list args )
{
  enum V3_IO_STATES state;
  u64  chars_output;
  u16  character;
  ascii * f_ascii, * s_ascii;
  unicode * f_unicode, * s_unicode;

  u32  tmp_u32;
  s32  tmp_s32;
  u64  tmp_u64;
  s64  tmp_s64;
  u128 tmp_u128;
  s128 tmp_s128;
  f64  tmp_f64;
  ascii * ptr_ascii;
  unicode * ptr_unicode;
  ascii tmp_ascii[48]; /* longest is s128? !!! floats may be longer? */
  u32 tmp_count;
  s32 tmp_sign;
  s32 i;
  u32 u;
  u8 * ptr_u8;
  static ascii hex_table[16] = { '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

  u32 done = V3_IO_NO;

  u32 width;
  u32 width_set;
  u32 pad_left;
  u32 pad_zero;
  u32 precision;
  u32 precision_set;

  _V3_SET64( chars_output, 00000000, 00000000 );

  /* check for valid parameters */
  if ( ( v3u64Eq( max_chars, chars_output ) ) || ( format == NULL ) )
    return( chars_output );

  f_ascii = (ascii *) format;
  f_unicode = (unicode *) format;
  s_ascii = (ascii *) string;
  s_unicode = (unicode *) string;

  _V3_IO_CLEARTMP
  _V3_IO_FETCHCHAR
  if ( character == 0 )
    return( chars_output );

  while ( done == V3_IO_NO )
  {
    switch ( state )
    {
      case V3_IO_TEXT:
        switch ( character )
        {
          case '%':
            state = V3_IO_PERCENT;
            break;
          case 0:
            return( chars_output );
          default:
            v3_PrintChar( file, string, char_width, character );
            _V3_IO_CHECKMAX
            break;
        }
        _V3_IO_FETCHCHAR
        break;
      case V3_IO_PERCENT:
        if ( character == '%' )
        {
          v3_PrintChar( file, string, char_width, character );
          _V3_IO_CHECKMAX
          _V3_IO_FETCHCHAR
          _V3_IO_CLEARTMP
        }
        else
        {
          if ( ( character >= '1' ) && ( character <= '9' ) )
          {
            /* We are defining the width */
            width_set = V3_IO_YES;
            state = V3_IO_WIDTH;
          }
          else
          {
            switch ( character )
            {
              case '0':
                pad_zero = V3_IO_YES;
                width_set = V3_IO_YES;
                state = V3_IO_WIDTH;
                break;
              case '-':
                pad_left = V3_IO_YES;
                width_set = V3_IO_YES;
                state = V3_IO_WIDTH;
                break;
              case '*':
                width = va_arg( args, u32 );
                if ( width > 0 ) /* avoid nasty bug */
                  width_set = V3_IO_YES;
                break;
              case '.':
                precision = 0;
                state = V3_IO_PREC;
                break;
              case 'c':
                state = V3_IO_SINGLE_CHAR;
                break;
              case 's':
                state = V3_IO_STRING;
                break;
              case 'b':
                state = V3_IO_BUFFER;
                break;
              case 'i':
                state = V3_IO_S32_DEC;
                break;
              case 'j':
                state = V3_IO_S64_DEC;
                break;
              case 'k':
                state = V3_IO_S128_DEC;
                break;
              case 'u':
               state = V3_IO_U32_DEC;
                break;
              case 'v':
                state = V3_IO_U64_DEC;
                break;
              case 'w':
                state = V3_IO_U128_DEC;
                break;
#if ( !defined( NO_IEEE_FLOAT ) )
              case 'f':
                state = V3_IO_F32;
                break;
              case 'g':
                state = V3_IO_F64;
                break;
              case 'F':
                state = V3_IO_F32_SCI;
                break;
              case 'G':
                state = V3_IO_F64_SCI;
                break;
#endif
              case 'X':
                state = V3_IO_U32_HEX;
                break;
              case 'Y':
                state = V3_IO_U64_HEX;
                break;
              case 'Z':
                state = V3_IO_U128_HEX;
                break;
              case 'p':
                state = V3_IO_POINTER;
                break;
              case 0:
                return( chars_output );
              default:
                state = V3_IO_TEXT;
                break;
            }
            _V3_IO_FETCHCHAR
          }
        }
        break;
      case V3_IO_WIDTH:
        switch ( character )
        {
          case '0':
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
            width = width * 10 + ( character - '0' );
            _V3_IO_FETCHCHAR
            break;
          case 0:
            return( chars_output );
          default:
            state = V3_IO_PERCENT;
            break;
         }
         break;
      case V3_IO_PREC:
        switch ( character )
        {
          case '*':
            precision = va_arg( args, u32 );
            precision_set = V3_IO_YES;
            _V3_IO_FETCHCHAR
            state = V3_IO_PERCENT;
            break;
          case '0':
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
            precision = precision * 10 + ( character - '0' );
            precision_set = V3_IO_YES;
            _V3_IO_FETCHCHAR
            break;
          case 0:
            return( chars_output );
          default:
            state = V3_IO_PERCENT;
            break;
        }
        break;
      case V3_IO_SINGLE_CHAR:
        if ( char_width == 1 )
        {
          v3_PrintChar( file, string, char_width, va_arg( args, ascii ) );
        }
        else
        {
          v3_PrintChar( file, string, char_width, va_arg( args, unicode ) );
        }
        _V3_IO_CHECKMAX
        _V3_IO_CLEARTMP
        break;
      case V3_IO_STRING:
        if ( char_width == 1 )
        {
          ptr_ascii = (ascii *) va_arg( args, ascii * );
          if ( precision_set )
          {
            if ( ptr_ascii != NULL )
            {
              u = 0;
              while ( ( u < precision ) && ( ptr_ascii[u] != 0 ) )
              {
                v3_PrintChar( file, string, char_width, ptr_ascii[u] );
                _V3_IO_CHECKMAX
                u++;
              }
            }
          }
        }
        else
        {
          ptr_unicode = (unicode *) va_arg( args, unicode * );
          if ( precision_set )
          {
            if ( ptr_unicode != NULL )
            {
              u = 0;
              while ( ( u < precision ) && ( ptr_unicode[u] != 0 ) )
              {
                v3_PrintChar( file, string, char_width, ptr_unicode[u] );
                _V3_IO_CHECKMAX
                u++;
              }
            }
          }
        }
        _V3_IO_CLEARTMP
        break;
      case V3_IO_BUFFER:
        ptr_u8 = va_arg( args, u8 * );
        if ( ( precision_set ) && ( ptr_u8 != NULL ) )
        {
          for ( u = 0 ; u < precision ; u++ )
          {
            v3_PrintChar( file, string, 1, ptr_u8[u] );
            _V3_IO_CHECKMAX
          }
        }
        _V3_IO_CLEARTMP
        break;
      case V3_IO_TMP_STRING:
        /*
          print tmp_string backwards
          using tmp_count, width, pad_left, pad_zero
        */
        if ( ( tmp_sign ) && ( width_set ) )
        {
          width--;
        }
        /* First pad as necessary (zero pad or left pad ) */
        if ( ( width_set ) && ( width > tmp_count ) &&
          ( pad_left == V3_IO_NO ) )
        {
          for ( u = 0 ; u < ( width - tmp_count ) ; u++ )
          {
            if ( pad_zero == V3_IO_YES )
            {
              v3_PrintChar( file, string, char_width, '0' );
            }
            else
            {
              v3_PrintChar( file, string, char_width, ' ' );
            }
            _V3_IO_CHECKMAX
          }
        }
        if ( tmp_sign )
        {
          v3_PrintChar( file, string, char_width, '-' );
          _V3_IO_CHECKMAX
        }
        for ( i = ( tmp_count - 1 ) ; i > -1 ; i-- )
        {
          v3_PrintChar( file, string, char_width, tmp_ascii[i] );
          _V3_IO_CHECKMAX
        }
        if ( ( width_set ) && ( width > tmp_count ) &&
          ( pad_left == V3_IO_YES ) )
        {
          for ( u = 0 ; u < ( width - tmp_count ) ; u++ )
          {
            v3_PrintChar( file, string, char_width, ' ' );
            _V3_IO_CHECKMAX
          }
        }
        _V3_IO_CLEARTMP
        break;
      case V3_IO_U32_DEC:
        tmp_u32 = va_arg( args, u32 );

        do
        {
          *ptr_ascii++ = (ascii) ( '0' + ( tmp_u32 % 10 ) );
          tmp_u32 = tmp_u32 / 10;
          tmp_count++;
        } while ( tmp_u32 > 0 );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_U64_DEC:
        tmp_u64 = va_arg( args, u64 );

        do
        {
          *ptr_ascii++ = (ascii) ( '0' + v3u32u64(
            v3u64Mod( tmp_u64, v3u64u32( 10 ) ) ) );
          tmp_u64 = v3u64Div( tmp_u64, v3u64u32( 10 ) );
          tmp_count++;
        } while ( v3u64Gt( tmp_u64, v3u64u32( 0 ) ) );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_U128_DEC:
        tmp_u128 = va_arg( args, u128 );

        do
        {
          *ptr_ascii++ = (ascii) ( '0' + v3u32u64( v3u64u128(
            v3u128Mod( tmp_u128, v3u128u32( (u32) 10 ) ) ) ) );
          tmp_u128 = v3u128Div( tmp_u128, v3u128u32( (u32) 10 ) );
          tmp_count++;
        } while ( v3u128Gt( tmp_u128, v3u128u32( (u32) 0 ) ) );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_S32_DEC:
        tmp_s32 = va_arg( args, s32 );

        if ( tmp_s32 < 0 )
        {
          tmp_sign = -1;
          tmp_s32 = -tmp_s32;
        }

        do
        {
          *ptr_ascii++ = (ascii) ( '0' + ( tmp_s32 % 10 ) );
          tmp_s32 = tmp_s32 / 10;
          tmp_count++;
        } while ( tmp_s32 > 0 );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_S64_DEC:
        tmp_s64 = va_arg( args, s64 );

        if ( v3s64Lt( tmp_s64, v3s64s32( 0 ) ) )
        {
          tmp_sign = -1;
          tmp_s64 = v3s64Sub( v3s64s32( 0 ), tmp_s64 );
        }

        do
        {
          *ptr_ascii++ = (ascii) ( '0' + v3s32s64(
            v3s64Mod( tmp_s64, v3s64s32( 10 ) ) ) );
          tmp_s64 = v3s64Div( tmp_s64, v3s64s32( 10 ) );
          tmp_count++;
        } while ( v3s64Gt( tmp_s64, v3s64s32( 0 ) ) );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_S128_DEC:
        tmp_s128 = va_arg( args, s128 );

        if ( v3s128Lt( tmp_s128, v3s128s32( 0 ) ) )
        {
          tmp_sign = -1;
          tmp_s128 = v3s128Sub( v3s128s32( 0 ), tmp_s128 );
        }

        do
        {
          *ptr_ascii++ = (ascii) ( '0' + v3s32s64( v3s64s128(
            v3s128Mod( tmp_s128, v3s128s32( 10 ) ) ) ) );
          tmp_s128 = v3s128Div( tmp_s128, v3s128s32( 10 ) );
          tmp_count++;
        } while ( v3s128Gt( tmp_s128, v3s128s32( 0 ) ) );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_U32_HEX:
       tmp_u32 = va_arg( args, u32 );

        do
        {
          *ptr_ascii++ = hex_table[ tmp_u32 & 0xF ];
          tmp_u32 = tmp_u32 >> 4;
          tmp_count++;
        } while ( tmp_u32 > 0 );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_U64_HEX:
        tmp_u64 = va_arg( args, u64 );

        do
        {
          *ptr_ascii++ = hex_table[ v3u32u64(
            v3u64And( tmp_u64, v3u64u32( 15 ) ) ) ];
          tmp_u64 = v3u64Rsh( tmp_u64, 4 );
          tmp_count++;
        } while ( v3u64Gt( tmp_u64, v3u64u32( 0 ) ) );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      case V3_IO_U128_HEX:
        tmp_u128 = va_arg( args, u128 );

        do
        {
          *ptr_ascii++ = hex_table[ v3u32u64( v3u64u128(
            v3u128And( tmp_u128, v3u128u32( (u32) 15 ) ) ) ) ];
          tmp_u128 = v3u128Rsh( tmp_u128, 4 );
          tmp_count++;
        } while ( v3u128Gt( tmp_u128, v3u128u32( 0 ) ) );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
#if ( !defined( NO_IEEE_FLOAT ) )
      case V3_IO_F32:
      case V3_IO_F64:
      {
        f64 whole;

        tmp_f64 = (f64) va_arg( args, f64 );

        if ( v3_isnan(tmp_f64) )
        {
          *ptr_ascii++ = (ascii) 'n';
          *ptr_ascii++ = (ascii) 'a';
          *ptr_ascii++ = (ascii) 'n';
          *ptr_ascii++ = 0;
          tmp_count+=3;
          state = V3_IO_TMP_STRING;
          break;
        }

        if ( tmp_f64 < 0 )
        {
          tmp_sign = -1;
          tmp_f64 = -tmp_f64;
        }

        if ( v3_isinf(tmp_f64) )
        {
          *ptr_ascii++ = (ascii) 'f';
          *ptr_ascii++ = (ascii) 'n';
          *ptr_ascii++ = (ascii) 'i';
          *ptr_ascii++ = 0;
          tmp_count+=3;
          state = V3_IO_TMP_STRING;
          break;
        }

        /* Get whole/fraction part using casts. */
        whole = floor(tmp_f64);
        tmp_f64 = tmp_f64 - whole;

        /* Set precision to "6" if not set (From cosmio.h)  */
        if ( precision_set )
        {
          if ( precision == 0 )
             tmp_f64 = 0.0;
        }
        else precision = 6;

        if ( tmp_f64 > 0 || precision )
        {
          for ( i = precision-1; i >= 0; i-- )
          {
            f64 tmp_whole;

            if (tmp_f64 > 0)
            {
              tmp_f64 = tmp_f64 * 10.0;
              tmp_whole = floor(tmp_f64);
              tmp_f64 = tmp_f64 - tmp_whole;

              /* round */
              if ( i == 0 && tmp_f64 > 0.5)
                 tmp_whole++;
            }
            else tmp_whole = 0;

            *(ptr_ascii+i) = (ascii) ( '0' + tmp_whole );
          }

          ptr_ascii += precision;
          tmp_count += precision;

          *ptr_ascii++ = (ascii) '.';
          tmp_count++;
        }

        /* print whole */
        do
        {
          *ptr_ascii++ = (ascii) ( '0' + fmod(whole,10.0) );
          whole = whole / 10;
          whole = floor(whole);
          tmp_count++;
        } while ( whole > 0 );

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      }
      case V3_IO_F32_SCI:
      case V3_IO_F64_SCI:
      {
        f64 whole;

        tmp_f64 = (f64) va_arg( args, f64 );

        if (v3_isnan(tmp_f64))
        {
          *ptr_ascii++ = (ascii) 'n';
          tmp_count++;
          *ptr_ascii++ = (ascii) 'a';
          tmp_count++;
          *ptr_ascii++ = (ascii) 'n';
          tmp_count++;
          *ptr_ascii++ = 0;
          state = V3_IO_TMP_STRING;
          break;
        }

        if ( tmp_f64 < 0 )
        {
          tmp_sign = -1;
          tmp_f64 = -tmp_f64;
        }

        if ( v3_isinf(tmp_f64) )
        {
          *ptr_ascii++ = (ascii) 'f';
          tmp_count++;
          *ptr_ascii++ = (ascii) 'n';
          tmp_count++;
          *ptr_ascii++ = (ascii) 'i';
          tmp_count++;
          *ptr_ascii++ = 0;
          state = V3_IO_TMP_STRING;
          break;
        }

        /* Count "E" digits */
        tmp_s32 = 0;
        while ( tmp_f64 < 1.0 )
        {
          tmp_s32--;
          tmp_f64 = tmp_f64 * 10.0;
        }
        while ( tmp_f64 >= 10.0 )
        {
          tmp_s32++;
          tmp_f64 = tmp_f64 / 10.0;
        }

        if (tmp_s32)
        {
          if (tmp_s32 < 0)
          {
               tmp_u32 = -tmp_s32;
          }
          else tmp_u32 = tmp_s32;

          do
          {
            *ptr_ascii++ = (ascii) ( '0' + ( tmp_u32 % 10 ) );
            tmp_u32 = tmp_u32 / 10;
            tmp_count++;
          } while ( tmp_u32 > 0 );

          if (tmp_s32 < 0)
          {
               *ptr_ascii++ = (ascii) '-';
               tmp_count++;
          }

          *ptr_ascii++ = (ascii) 'E';
          tmp_count++;
        }

        whole = floor(tmp_f64);
        tmp_f64 = tmp_f64 - whole;

        /* Set precision to "6" if not set (From cosmio.h)  */
        if ( precision_set )
        {
          if ( precision == 0 )
             tmp_f64 = 0.0;
        }
        else precision = 6;

        for ( i = precision-1; i >= 0; i-- )
        {
         f64 tmp_whole;

         if (tmp_f64 > 0)
         {
           tmp_f64 = tmp_f64 * 10.0;
           tmp_whole = floor(tmp_f64);
           tmp_f64 = tmp_f64 - tmp_whole;

           /* round */
           if ( i == 0 && tmp_f64 > 0.5)
              tmp_whole++;
         }
         else tmp_whole = 0;

         *(ptr_ascii+i) = (ascii) ( '0' + tmp_whole );
        }
        ptr_ascii += precision;
        tmp_count += precision;

        *ptr_ascii++ = (ascii) '.';
        *ptr_ascii++ = (ascii) ( '0' + whole );
        tmp_count += 2;

        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
      }
#endif
      case V3_IO_POINTER:
        switch ( sizeof( void * ) )
        {
          case 4:
            tmp_u32 = (u32) va_arg( args, u32 );

            for ( u = 0 ; u < 8 ; u++ )
            {
              *ptr_ascii++ = hex_table[ tmp_u32 & 0x0F ];
              tmp_u32 = tmp_u32 >> 4;
              tmp_count++;
            }
            break;
          case 8:
            /*
              this will be wrong on little endian system with
              fake u64's and 64bit pointers, unlikely to exist.
            */
            tmp_u64 = va_arg( args, u64 );

            for ( u = 0 ; u < 16 ; u++ )
            {
              *ptr_ascii++ = hex_table[ v3u32u64( v3u64Mod( tmp_u64,
                v3u64u32( 16 ) ) ) ];
              tmp_u64 = v3u64Rsh( tmp_u64, 4 );
              tmp_count++;
            }
            break;
          case 2:
            tmp_u32 = (u32) va_arg( args, u32 );

            for ( u = 0 ; u < 4 ; u++ )
            {
              *ptr_ascii++ = hex_table[ tmp_u32 & 0x0F ];
              tmp_u32 = tmp_u32 >> 4;
              tmp_count++;
            }
            break;
        }
        *ptr_ascii++ = 0;
        state = V3_IO_TMP_STRING;
        break;
    }
  }
  return( v3u64u32( 0 ) );
}

s32 v3_PrintChar( v3_FILE * file, void * string_ptr, u8 char_width,
  u16 character )
{
  ascii   char_ascii;
  unicode char_unicode;
  u8 array[2];

  if ( string_ptr != NULL )
  {
    if ( char_width == 1 )
    {
      *( (ascii *) string_ptr ) = (ascii) character;
    }
    else
    {
      *( (unicode *) string_ptr ) = (unicode) character;
    }
  }
  else if ( file != NULL )
  {
    if ( char_width == 1 )
    {
      char_ascii = (ascii) character;
      if ( !v3u64Eq( v3FileWrite( file, &char_ascii, v3u64u32( 1 ) ),
                    v3u64u32( 1 ) ) )
      {
        return( V3_FAIL );
      }
    }
    else
    {
      char_unicode = (unicode) character;
      v3u16Save( array, &char_unicode ); /* important */
      if ( !v3u64Eq( v3FileWrite( file, array, v3u64u32( 2 ) ),
                    v3u64u32( 2 ) ) )
      {
        return( V3_FAIL );
      }
    }
  }
  else /* standard output */
  {
    if ( char_width == 1 )
    {
      if ( putchar( (ascii) character ) == EOF )
      {
        return( V3_FAIL );
      }
    }
    else
    {
      /* !!! how do we do this? */
      if ( putchar( (unicode) character ) == EOF )
      {
        return( V3_FAIL );
      }
    }
  }

  return( V3_PASS );
}

/* Is f64 NAN? */
s32 v3_isnan( f64 number )
{
  u32 * n;

  n  = (u32 *) &number;

  return( ( ( n[0] & 0x7FF00000 ) == 0x7FF00000 )
            && ( ( ( 0x000FFFFF & n[0] ) | n[1] ) != 0 ) );
}

/* Is f64 INF? */
s32 v3_isinf( f64 number )
{
  u32 * n;

  n  = (u32 *) &number;

  return( ( ( n[0] & 0x7FF00000 ) == 0x7FF00000 )
            && ( ( ( 0x000FFFFF & n[0] ) | n[1] ) == 0 ) );
}

/* testing */

typedef struct
{
  f64 data;
  const ascii * format;
  const ascii * result;
} v3_FLOAT_TEST;

/* tests for %f, %g, %F, %G */
v3_FLOAT_TEST float_tests[] =
{
/*  { (f64) 1.0, (ascii *) "", (ascii *) "" }, */
  { (f64) 1.0, (ascii *) "%.1f", (ascii *) "1.0" },
  { (f64) -1.0, (ascii *) "%.1f", (ascii *) "-1.0" },
  { (f64) 0.0, (ascii *) NULL, (ascii *) NULL }
};

typedef struct
{
  u32 data;
  const ascii * format;
  const ascii * result;
} v3_SU32_TEST;

/* tests for %i, %u, %X */
v3_SU32_TEST su32_tests[] =
{
/*  { (u32) 0x00000000, (ascii *) "", (ascii *) "" }, */
  { (u32) 0x00000001, (ascii *) "%u", (ascii *) "1" },
  { (u32) 0x00000001, (ascii *) "%i", (ascii *) "1" },
  { (u32) 0xFFFFFFFF, (ascii *) "%u", (ascii *) "4294967295" },
  { (u32) 0xFFFFFFFF, (ascii *) "%i", (ascii *) "-1" },
  { (u32) 0x00000001, (ascii *) "%X", (ascii *) "1" },
  { (u32) 0x00000001, (ascii *) "%04X", (ascii *) "0001" },
  { (u32) 0xFFFFFFFF, (ascii *) "%X", (ascii *) "FFFFFFFF" },
  { (u32) 0.0, (ascii *) NULL, (ascii *) NULL }
};

/* testing */

s32 v3_TestCosmIO( void )
{
  ascii a_str1[1024] = "26 character test string.";
  ascii a_str2[512] = "";
  u32 i;
  ascii * found;

  u32 tmpNum;
  ascii ascNum[6] = " 0xfF";
  /* L" 0xfF" in several languages */
  unicode uniNum[6] = { 0x1680, '0', 'x', 0x0396, 0x05D5 , 0 };

  /* ASCII tests */

  if ( !v3u64Eq( v3StrLengthA( NULL ), v3u64u32( 0 ) ) )
  {
    return( -1 );
  }

  if ( !v3u64Eq( v3StrLengthA( a_str1 ), v3u64u32( 25 ) ) )
  {
    return( -2 );
  }

  /* v3StrCopyA, Copy a_str1 to a_str2 then compare. */
  if ( v3StrCopyA( a_str2, a_str1, v3u64u32( 512 ) ) != V3_PASS )
  {
    return( -3 );
  }

  for ( i = 0 ; i < 26 ; i++ )
  {
    if ( a_str1[i] != a_str2[i] )
    {
      return( -4 );
    }
  }

  /* Compare a_str1 with a_str2. a_str1 is the same as a_str2 as a result
     of the above v3StrCopyA test. Should return 0. */
  if ( v3StrCmpA( a_str1, a_str2, v3u64u32( sizeof ( a_str1 ) ) ) != 0 )
  {
    return( -5 );
  }

  a_str2[3] = 'b';

  /* Should return a positive number */
  if ( v3StrCmpA( a_str1, a_str2, v3u64u32( sizeof ( a_str1 ) ) ) <= 0 )
  {
    return( -6 );
  }

  /* Should return a negative number */
  if ( v3StrCmpA( a_str2, a_str1, v3u64u32( sizeof ( a_str1 ) ) ) >= 0 )
  {
    return( -7 );
  }

  /* Find character 'a' in a_str1 */
  if ( ( found = v3StrCharA( a_str1, 'a', v3u64u32( sizeof ( a_str1 ) ) ) )
    == NULL )
  {
    return( -8 );
  }

  /* Convert a string containing " 0xfF" to a number */
  tmpNum = 0;
  v3u32A( &tmpNum, 0, ascNum, 0 );
  if ( tmpNum != 255 )
  {
    return( -9 );
  }
  /* Unicode tests */
  tmpNum = 0;
  v3u32U( &tmpNum, 0, uniNum, 0 );
  if ( tmpNum != 255 )
  {
    return( -10 );
  }

  return( V3_PASS );
}
