/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_EMAIL_H
#define V3_EMAIL_H

#include "cputypes.h"
#include "cosmnet.h"

#define V3_EMAIL_ERROR_NONE     0
#define V3_EMAIL_ERROR_TO       -1 /* to rejected */
#define V3_EMAIL_ERROR_FROM     -2 /* from rejected */
#define V3_EMAIL_ERROR_HOST     -3 /* Invalid SMTP host */
#define V3_EMAIL_ERROR_MESSAGE  -4 /* message == NULL / length != 0 */
#define V3_EMAIL_ERROR_NORELAY  -5 /* Relay not allowed */
#define V3_EMAIL_ERROR_ABORTED  -6 /* Unable to complete transmission */

/* High Level Functions */

s32 v3EmailSMTP( const u32 smtp_server, const ascii * to, const ascii * from,
  const ascii * subject, const ascii * message, u32 length );
  /*
    Send SMTP email message of length length with subject using the given
    smtp_server host. The To: and From: fields will be set to to and from.
    SMTP is a 7-bit protocol so only plain text can be sent.
    Returns: V3_PASS on success, or an error code on failure.
  */

/* Low Level Functions */

s32 v3_EmailGetReply( ascii * reply, v3_NET * net );
  /*
    Get the complete reply message and only return the numeric part, setting
    reply to a three character string. reply must be at least 4 characters
    long.
    Returns: V3_PASS on success or V3_FAIL on failure.
  */

/* testing */

s32 v3_TestEmail( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
