/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_LOG_H
#define V3_LOG_H

#include "cputypes.h"
#include "cosmfile.h"
#include "cosmtask.h"

#define V3_LOG_STATUS_NULL  0
#define V3_LOG_STATUS_INIT  1

#define V3_LOG_MODE_NUMBER  1
#define V3_LOG_MODE_BITS    2

#define V3_LOG_ECHO    3
#define V3_LOG_NOECHO  6

#define V3_LOG_ERROR_NONE    0
#define V3_LOG_ERROR_MODE    -1  /* Invalid mode */
#define V3_LOG_ERROR_NAME    -2  /* Invalid filename */
#define V3_LOG_ERROR_INIT    -3  /* Log not initialized */
#define V3_LOG_ERROR_ACCESS  -4  /* File access error */

typedef struct
{
  v3_FILENAME filename;
  v3_FILE file;
  u32 status;
  u32 mode;
  u32 level;
  v3_MUTEX lock;
} v3_LOG;

s32 v3LogOpen( v3_LOG * log, ascii * filename, u32 max_level, u32 mode );
  /*
    Initialize the log, setting the filename, max_level, and mode.
    If V3_LOG_MODE_NUMBER then log any message with an equal or lower level
    than max_. If V3_LOG_MODE_BITS then log any message with it's level
    matching a bit in max_level. Initializing a log with a NULL filename or
    invalid mode causes failure.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3Log( v3_LOG * log, u32 level, u32 echo, const ascii * format, ... );
  /*
    Write the formatted string to the log. (See v3PrintA for format)
    String will be logged only if the level is acceptable to the open log.
    Level 0 will always logged. If echo is V3_LOG_ECHO and the level passes
    then the string will also be printed to the standard output device.
    Otherwise if level does not pass or echo is V3_LOG_NOECHO nothing will be
    printed.
    Returns: V3_PASS on success, or an error code on failure
  */

s32 v3LogClose( v3_LOG * log );
  /*
    Close the file.
    Returns: V3_PASS on success, or an error code on failure
  */



/* testing */

s32 v3_TestLog( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
